from alembic.operations.ops import CreateTableOp, DropTableOp, UpgradeOps
from alembic.runtime.environment import MigrationContext
from nose.tools import *
from retort.migration import DROP_OPERATIONS, Migration
from sqlalchemy import Column, Integer, MetaData, String
from sqlalchemy.ext.declarative import declarative_base

from .target_databases import Target
from .util import captured_output


def setup():
    Target.drop_and_create_database()


class TestGetMigrationContext:
    @classmethod
    def metadata(cls):
        Base = declarative_base()
        return Base.metadata

    def test(self):
        engine = Target.get_engine()
        migration = Migration(engine, self.metadata)
        context = migration.get_migration_context()
        assert_is_instance(context, MigrationContext)

    def test_as_sql(self):
        engine = Target.get_engine()
        migration = Migration(engine, self.metadata)
        context = migration.get_migration_context(as_sql=True)
        assert_is_instance(context, MigrationContext)
        assert_true(context.as_sql)


class TestMigrationOps:
    def create_table(self):
        Base = declarative_base()
        engine = Target.get_engine()

        class ExistingFooBar(Base):
            __tablename__ = 'foobar'
            id = Column(Integer, primary_key=True)
            name = Column(String(10))
            foo = Column(String(10))
            xxx = Column(String(10))
        Base.metadata.drop_all(bind=engine)
        Base.metadata.create_all(bind=engine)

    def setup(self):
        Target.drop_all_table()
        self.Base = declarative_base()
        self.engine = Target.get_engine()

        class FooBar(self.Base):
            __tablename__ = 'foobar'
            id = Column(Integer, primary_key=True)
            name = Column(String(10))
            yyy = Column(String(10))

    def is_included_operations(self, migration_ops, expected_operations):
        if not hasattr(migration_ops, 'ops') or len(migration_ops.ops) is 0:
            return isinstance(migration_ops, expected_operations)
        return any(self.is_included_operations(ops, expected_operations) for ops in migration_ops.ops)

    def is_included_drop_operations(self, migration_ops):
        return self.is_included_operations(migration_ops, DROP_OPERATIONS)

    def test_create_table(self):
        migration = Migration(self.engine, self.Base.metadata)
        migration_ops = migration.migration_ops
        assert_is_instance(migration_ops, UpgradeOps)
        ok_(len(migration_ops.ops) == 1)
        assert_is_instance(migration_ops.ops[0], CreateTableOp)

    def test_drop_table(self):
        self.create_table()
        base_without_table = declarative_base()
        migration = Migration(self.engine, base_without_table.metadata)
        migration_ops = migration.migration_ops
        assert_is_instance(migration_ops, UpgradeOps)
        ok_(len(migration_ops.ops) == 1)
        assert_is_instance(migration_ops.ops[0], DropTableOp)

    def test_alter_table(self):
        self.create_table()
        migration = Migration(self.engine, self.Base.metadata)
        migration_ops = migration.migration_ops
        assert_is_instance(migration_ops, UpgradeOps)
        ok_(self.is_included_drop_operations(migration_ops))

    def test_alter_without_drop_operation(self):
        self.create_table()
        migration = Migration(self.engine, self.Base.metadata, without_drop=True)
        migration_ops = migration.migration_ops
        assert_is_instance(migration_ops, UpgradeOps)
        ok_(not self.is_included_drop_operations(migration_ops))


class TestApply:
    def create_table(self):
        Base = declarative_base()
        engine = Target.get_engine()

        class ExistingFooBar(Base):
            __tablename__ = 'foobar'
            id = Column(Integer, primary_key=True)
            name = Column(String(10))
            foo = Column(String(10))
            xxx = Column(String(10))
        Base.metadata.drop_all(bind=engine)
        Base.metadata.create_all(bind=engine)

    def setup(self):
        Target.drop_all_table()
        self.create_table()
        self.Base = declarative_base()
        self.engine = Target.get_engine()

        class FooBar(self.Base):
            __tablename__ = 'foobar'
            id = Column(Integer, primary_key=True)
            name = Column(String(10))
            yyy = Column(String(10))

    def teardown(self):
        Target.drop_all_table()

    def test(self):
        migration = Migration(self.engine, self.Base.metadata)
        migration.apply()
        meta = MetaData()
        meta.reflect(bind=self.engine)
        foobar_table = meta.tables['foobar']
        assert_is_instance(foobar_table.columns['id'].type, Integer)
        assert_is_instance(foobar_table.columns['name'].type, String)
        assert_is_instance(foobar_table.columns['yyy'].type, String)
        assert_raises(KeyError, lambda: foobar_table.columns['foo'])
        assert_raises(KeyError, lambda: foobar_table.columns['xxx'])


class TestPrintEngineInfo:
    @classmethod
    def metadata(cls):
        Base = declarative_base()
        return Base.metadata

    def test(self):
        engine = Target.get_engine()
        migration = Migration(engine, self.metadata)
        with captured_output() as (out, _err):
            migration.print_engine_info()
        outputs = out.getvalue().split('\n')
        eq_(len(outputs), 3)
        eq_(outputs[0], 'url: {}'.format(engine.url))
        eq_(outputs[1], 'logging_name: None')
        eq_(outputs[2], '')

    def test_with_logging_name(self):
        logging_name = 'testing'
        engine = Target.get_engine(logging_name=logging_name)
        migration = Migration(engine, self.metadata)
        with captured_output() as (out, _err):
            migration.print_engine_info()
        outputs = out.getvalue().split('\n')
        eq_(len(outputs), 3)
        eq_(outputs[0], 'url: {}'.format(engine.url))
        eq_(outputs[1], 'logging_name: testing')
        eq_(outputs[2], '')



import os
import sys
from distutils.util import strtobool

from .constant import CONFIG_FILE_NAME
from .migration import Migration


def prompt(query):
    sys.stdout.write('%s [y/n]: ' % query)
    val = input()
    try:
        ret = strtobool(val)
    except ValueError:
        print('Please answer with a y/n\n')
        return prompt(query)
    return ret


def get_config():
    sys.path.append(os.getcwd())
    from importlib.machinery import SourceFileLoader
    retort_conf = SourceFileLoader('retort_conf', CONFIG_FILE_NAME).load_module()
    return retort_conf


def get_migrations(without_drop=False):
    config = get_config()
    return [Migration(target['engine'], target['metadata'], without_drop=without_drop) for target in config.TARGETS]


def print_codes(migrations, sql=False):
    for m in migrations:
        print_code(m, sql)


def print_code(migration, sql=False):
    print('====================')
    migration.print_engine_info()
    print('====================')
    migration.print_code(sql)



import os
import shutil
import sys
import time
from argparse import ArgumentParser

from . import cmd_helper, __version__


def main(args):
    pass


def init(args):
    stub_file_path = os.path.join(os.path.dirname(__file__), 'data', 'retort_conf_example.py.txt')
    if os.path.exists('./retort_conf.py'):
        print('retort_conf.py already exists.')
        return
    shutil.copyfile(stub_file_path, 'retort_conf.py')
    print('Create retort_conf.py.')


def apply(args):
    migrations = cmd_helper.get_migrations(args.without_drop)

    if not args.yes:
        cmd_helper.print_codes(migrations, args.sql)
        if not cmd_helper.prompt('Do you really want to apply this?'):
            print('Cancelled.')
            return
    print('')
    print('Applying migration......')
    for m in migrations:
        cmd_helper.print_code(m, args.sql)
        if not args.dry_run:
            t1 = time.time()
            m.apply()
            t2 = time.time()
            print('---> Processing time: ' + str(round((t2 - t1), 4)) + '(sec)')
    print('Complete!')


def print_operations(args):
    migrations = cmd_helper.get_migrations(args.without_drop)
    cmd_helper.print_codes(migrations, args.sql)


def command():
    parser = ArgumentParser(description='Retort v{}: schema migration tool for SQLAlchemy'.format(__version__))
    parser.set_defaults(func=main)
    parser.add_argument('--version', action='version', version=__version__)
    subparsers = parser.add_subparsers()

    subparser_init = subparsers.add_parser('init')
    subparser_init.set_defaults(func=init)

    subparser_apply = subparsers.add_parser('apply')
    subparser_apply.add_argument('--sql', action='store_const', const=True, default=False)
    subparser_apply.add_argument('--dry-run', action='store_const', const=True, default=False)
    subparser_apply.add_argument('--yes', action='store_const', const=True, default=False)
    subparser_apply.add_argument('--without-drop', action='store_const', const=True, default=False)
    subparser_apply.set_defaults(func=apply)

    subparser_print_operations = subparsers.add_parser('print_operations')
    subparser_print_operations.add_argument('--sql', action='store_const', const=True, default=False)
    subparser_print_operations.add_argument('--without-drop', action='store_const', const=True, default=False)
    subparser_print_operations.set_defaults(func=print_operations)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()
    args.func(args)



from alembic.operations.ops import DropColumnOp, DropConstraintOp, DropIndexOp, DropTableOp

CONFIG_FILE_NAME = 'retort_conf.py'
DROP_OPERATIONS = (DropTableOp, DropColumnOp, DropIndexOp, DropConstraintOp)
DEFAULT_OPTS = {'compare_type': True, 'compare_server_default': True}



from invoke import task
from retort import __version__


@task
def build(ctx):
    ctx.run('python setup.py sdist bdist_wheel')


@task
def upload_pypi(ctx, _all=False, testpypi=False, dry_run=False):
    if _all:
        command = 'twine upload dist/*'
    else:
        command = 'twine upload dist/retort-{}*'.format(__version__)
    if testpypi:
        command = command + ' -r testpypi'

    if dry_run:
        print(command)
    else:
        ctx.run(command)



import os
from abc import ABCMeta, abstractmethod

from sqlalchemy import MetaData, create_engine

USER = os.environ.get('DB_USER', 'retort_test_user')
PASSWORD = os.environ.get('DB_PASSWORD', 'retort_test_user')
HOST = os.environ.get('DB_HOST', 'localhost')
NAME = os.environ.get('DB_NAME', 'retort_test_db')


class AbstractDatabase(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def drop_and_create_database(cls):
        pass

    @classmethod
    @abstractmethod
    def get_engine(cls, logging_name=None):
        pass

    @classmethod
    def drop_all_table(cls):
        meta = MetaData()
        engine = cls.get_engine()
        meta.reflect(bind=engine)
        meta.drop_all(engine)
        engine.dispose()

    @classmethod
    def get_url(cls, without_dbname=False):
        if without_dbname:
            return '{}+{}://{}:{}@{}'.format(cls.DIALECT, cls.DRIVER, USER, PASSWORD, HOST)
        else:
            return '{}+{}://{}:{}@{}/{}'.format(cls.DIALECT, cls.DRIVER, USER, PASSWORD, HOST, NAME)


class MySQL(AbstractDatabase):
    DIALECT = 'mysql'
    DRIVER = 'pymysql'

    @classmethod
    def drop_and_create_database(cls):
        engine = create_engine(cls.get_url(without_dbname=True))
        drop_str = 'DROP DATABASE IF EXISTS {};'.format(NAME)
        create_str = 'CREATE DATABASE {};'.format(NAME)
        engine.execute(drop_str)
        engine.execute(create_str)
        engine.dispose()

    @classmethod
    def get_engine(cls, logging_name=None):
        return create_engine(cls.get_url(), logging_name=logging_name)


class PostgreSQL(AbstractDatabase):
    DIALECT = 'postgresql'
    DRIVER = 'psycopg2'

    @classmethod
    def drop_and_create_database(cls):
        engine = create_engine(cls.get_url(without_dbname=True), isolation_level='AUTOCOMMIT')
        drop_str = 'DROP DATABASE IF EXISTS {};'.format(NAME)
        create_str = 'CREATE DATABASE {};'.format(NAME)
        engine.execute(drop_str)
        engine.execute(create_str)
        engine.dispose()

    @classmethod
    def get_engine(cls, logging_name=None):
        return create_engine(cls.get_url(), logging_name=logging_name, isolation_level='AUTOCOMMIT')


_target_diarect = os.environ.get('TARGET_DIARECT', 'mysql')

if _target_diarect == 'mysql':
    Target = MySQL
elif _target_diarect == 'postgresql':
    Target = PostgreSQL
else:
    Target = None



import importlib
import re

import autopep8
import sqlalchemy
from alembic.autogenerate.api import AutogenContext, produce_migrations, render
from alembic.operations import Operations
from alembic.runtime.environment import MigrationContext

from .constant import DEFAULT_OPTS, DROP_OPERATIONS
from .error import MigrationError


class Migration:
    def __init__(self, engine, metadata, without_drop=False, opts=None):
        self.engine = engine
        self.metadata = metadata
        self._without_drop = without_drop
        self._opts = DEFAULT_OPTS.copy()
        if opts:
            self._opts.update(opts)
        self._is_applied = False
        self._migration_ops = None
        self._autogen_context = None

    @property
    def is_applied(self):
        return self._is_applied

    def get_migration_context(self, as_sql=False):
        opts = self._opts.copy()
        opts['as_sql'] = as_sql
        return MigrationContext.configure(self.engine.connect(), opts=opts)

    @property
    def migration_ops(self):
        if not self._migration_ops:
            migration_context = self.get_migration_context()
            self._migration_ops = produce_migrations(migration_context, self.metadata).upgrade_ops
            if self._without_drop:
                self._remove_drop_ops(self._migration_ops)
        return self._migration_ops

    def _remove_drop_ops(self, migration_ops):
        if hasattr(migration_ops, 'ops'):
            migration_ops.ops = [x for x in migration_ops.ops if not isinstance(x, DROP_OPERATIONS)]
            for ops in migration_ops.ops:
                self._remove_drop_ops(ops)

    @property
    def autogen_context(self):
        if not self._autogen_context:
            opts = {
                'sqlalchemy_module_prefix': 'sa.',
                'alembic_module_prefix': 'op.',
                'render_item': None,
                'render_as_batch': False,
            }
            self._autogen_context = AutogenContext(self.get_migration_context(), opts=opts)
        return self._autogen_context

    def get_code(self):
        python_code = render._render_cmd_body(self.migration_ops, self.autogen_context)
        code_without_comment_and_indent = '\n'.join(
            [x for x in python_code.split('\n') if not x.startswith('#')])
        return autopep8.fix_code(code_without_comment_and_indent)

    def get_globals(self, op):
        dialect = self.autogen_context.dialect
        _globals = {
            'op': op,
            'sa': sqlalchemy,
        }
        matching_result = re.match(r'sqlalchemy\.dialects\.(\w+)', type(dialect).__module__)
        if matching_result:
            modname = matching_result.group(0)
            _globals.update({type(dialect).name: importlib.import_module(modname)})
        return _globals

    def apply(self):
        if self.is_applied:
            raise MigrationError
        with Operations.context(self.get_migration_context()) as op:
            python_code = self.get_code()
            exec(python_code, self.get_globals(op))
        self._is_applied = True

    def print_engine_info(self):
        print('\n'.join([
            'url: ' + str(self.engine.url),
            'logging_name: ' + str(self.engine.logging_name),
        ]))

    def print_code(self, sql=False):
        if sql:
            with Operations.context(self.get_migration_context(as_sql=True)) as op:
                python_code = self.get_code()
                exec(python_code, self.get_globals(op))
        else:
            print(self.get_code())



from os import path

from retort import __version__
from setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

tests_require = [name.rstrip() for name in open('requirements.txt').readlines()]


setup(
    name='retort',
    version=__version__,
    python_requires='>=3.3',
    description='Retort is a schema migration tool for SQLAlchemy.',
    long_description=long_description,
    license='BSD License (2-Clause)',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Database :: Front-Ends',
    ],
    keywords='SQLAlchemy migrations',
    author='NAKAMORI Ryosuke',
    author_email='me@tpdn.kim',
    url='https://github.com/tpdn/retort',
    packages=['retort'],
    package_dir={'retort': 'retort'},
    package_data={'retort': ['data/retort_conf_example.py.txt']},
    install_requires=[
        'SQLAlchemy', 'alembic', 'autopep8'
    ],
    tests_require=tests_require,
    test_suite='nose.collector',
    entry_points={
        'console_scripts': 'retort = retort.command:command'
    },
)
